library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.generic_type_pkg.all;

entity dptc is
  generic (
    Nv : integer := 2;
    Nh : integer := 2;
    N  : integer := 2;
    CUSTOM_WIDTH : integer := 16
  );
  port (
    clk       : in std_logic;
    reset_n   : in std_logic;
    enable    : in std_logic;
    x_matrix  : in my_custom_matrix(0 to Nv-1, 0 to N-1); 
    y_matrix  : in my_custom_matrix(0 to Nh-1, 0 to N-1);
    out_valid : out std_logic;                           
    result_matrix : out dot_matrix_type(0 to Nv-1, 0 to Nh-1) 
  );
end entity;

architecture Behavioral of dptc is

  component ddot_unit
    generic (
      N : integer := 2;
      CUSTOM_WIDTH : integer := 16
    );
    port (
      clk       : in std_logic;
      reset_n   : in std_logic;
      enable    : in std_logic;
      x_vec     : in my_custom_vector(0 to N-1);
      y_vec     : in my_custom_vector(0 to N-1);
      out_valid : out std_logic;
      dot_out   : out dot_product_type
    );
  end component;

  signal ddot_out : dot_matrix_type(0 to Nv-1, 0 to Nh-1);
  signal valid_signals : std_logic_vector(0 to (Nv*Nh)-1);
  signal any_valid : std_logic;

  type my_vector_array is array (natural range <>) of my_custom_vector(0 to N-1);
  signal x_vecs : my_vector_array(0 to Nv-1);
  signal y_vecs : my_vector_array(0 to Nh-1);

begin

  -- Extract row and column vectors
  process(x_matrix, y_matrix)
  begin
    for i in 0 to Nv-1 loop
      for k in 0 to N-1 loop
        x_vecs(i)(k) <= x_matrix(i, k);
      end loop;
    end loop;

    for j in 0 to Nh-1 loop
      for k in 0 to N-1 loop
        y_vecs(j)(k) <= y_matrix(j, k);
      end loop;
    end loop;
  end process;

  -- Generate dot units
  gen_rows: for i in 0 to Nv-1 generate
    gen_cols: for j in 0 to Nh-1 generate
      ddot_inst : ddot_unit
        generic map (
          N => N,
          CUSTOM_WIDTH => CUSTOM_WIDTH
        )
        port map (
          clk       => clk,
          reset_n   => reset_n,
          enable    => enable,
          x_vec     => x_vecs(i),
          y_vec     => y_vecs(j),
          out_valid => valid_signals(i*Nh + j),
          dot_out   => ddot_out(i,j)
        );
    end generate;
  end generate;

  -- Connect output matrix
  result_matrix <= ddot_out;

  -- One-valid output logic
  process(valid_signals)
  begin
    any_valid <= '0';
    for i in valid_signals'range loop
      if valid_signals(i) = '1' then
        any_valid <= '1';
      end if;
    end loop;
  end process;

  out_valid <= any_valid;

end architecture;
