library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.generic_type_pkg.all;

entity tile_unit is
  generic (
    N  : integer := 2; 
    Nv : integer := 2;
    Nh : integer := 2;
    Nm : integer := 4;
    Nd : integer := 2;  -- Number of DPTC cores
    CUSTOM_WIDTH : integer := 16
  );
  port (
    clk        : in std_logic;
    reset_n    : in std_logic;
    enable     : in std_logic;
    m1_chunk   : in my_custom_matrix(0 to Nv-1, 0 to Nm-1); 
    m2_chunk   : in my_custom_matrix(0 to Nh-1, 0 to Nm-1);  
    out_valid  : out std_logic;
    result_out : out dot_matrix_type(0 to Nv-1, 0 to Nh-1)
  );
end entity;

architecture Behavioral of tile_unit is

  component dptc
    generic (
      Nv : integer := 2;
      Nh : integer := 2;
      N  : integer := 2;
      CUSTOM_WIDTH : integer := 16
    );
    port (
      clk           : in std_logic;
      reset_n       : in std_logic;
      enable        : in std_logic;
      x_matrix      : in my_custom_matrix(0 to Nv-1, 0 to N-1);
      y_matrix      : in my_custom_matrix(0 to Nh-1, 0 to N-1);
      out_valid     : out std_logic;
      result_matrix : out dot_matrix_type(0 to Nv-1, 0 to Nh-1)
    );
  end component;

  -- Output from DPTCs
  type matrix_array is array(0 to Nd-1) of dot_matrix_type(0 to Nv-1, 0 to Nh-1);
  signal dptc_outputs : matrix_array;

  signal valid_signals : std_logic_vector(0 to Nd-1);

  -- Sliced chunks for each DPTC
  type chunk_matrix_array is array(0 to Nd-1) of my_custom_matrix(0 to Nv-1, 0 to N-1);
  signal m1_slices : chunk_matrix_array;
  signal m2_slices : chunk_matrix_array;

  -- Accumulation registers
  signal sum_result_reg : dot_matrix_type(0 to Nv-1, 0 to Nh-1);
  signal sum_valid_reg  : std_logic := '0';

begin

  -- Slice m1_chunk and m2_chunk for each DPTC
  slicer_proc: process(m1_chunk, m2_chunk)
  begin
    for d in 0 to Nd-1 loop
      for i in 0 to Nv-1 loop
        for k in 0 to N-1 loop
          m1_slices(d)(i,k) <= m1_chunk(i, d*N + k);
        end loop;
      end loop;

      for j in 0 to Nh-1 loop
        for k in 0 to N-1 loop
          m2_slices(d)(j,k) <= m2_chunk(j, d*N + k);
        end loop;
      end loop;
    end loop;
  end process;

  -- Instantiate DPTC cores
  dptc_gen: for d in 0 to Nd-1 generate
    dptc_inst : dptc
      generic map (
        Nv => Nv,
        Nh => Nh,
        N  => N,
        CUSTOM_WIDTH => CUSTOM_WIDTH
      )
      port map (
        clk           => clk,
        reset_n       => reset_n,
        enable        => enable,
        x_matrix      => m1_slices(d),
        y_matrix      => m2_slices(d),
        out_valid     => valid_signals(d),
        result_matrix => dptc_outputs(d)
      );
  end generate;

  -- Sum the results of all DPTC cores
  sum_proc: process(clk, reset_n)
    variable acc       : dot_matrix_type(0 to Nv-1, 0 to Nh-1);
    variable all_valid : std_logic;
  begin
    if reset_n = '0' then
      sum_result_reg <= (others => (others => (others => '0')));
      sum_valid_reg  <= '0';
    elsif rising_edge(clk) then
      all_valid := '1';
      for d in 0 to Nd-1 loop
        if valid_signals(d) /= '1' then
          all_valid := '0';
        end if;
      end loop;

      if all_valid = '1' then
        -- Clear accumulator
        for i in 0 to Nv-1 loop
          for j in 0 to Nh-1 loop
            acc(i,j) := (others => '0');
          end loop;
        end loop;

        -- Accumulate DPTC results
        for d in 0 to Nd-1 loop
          for i in 0 to Nv-1 loop
            for j in 0 to Nh-1 loop
              acc(i,j) := resize(acc(i,j), acc(i,j)'length) + resize(dptc_outputs(d)(i,j), acc(i,j)'length);
            end loop;
          end loop;
        end loop;

        sum_result_reg <= acc;
        sum_valid_reg  <= '1';
      else
        sum_valid_reg <= '0';
      end if;
    end if;
  end process;

  result_out <= sum_result_reg;
  out_valid  <= sum_valid_reg;

end architecture;
