library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.generic_type_pkg.all;

entity tile_array is
  generic (
    Np : integer := 4;
    Nm : integer := 4;
    Nq : integer := 4;
    Nt : integer := 2;
    Nd : integer := 2;
    Nh : integer := 2;
    N  : integer := 2 
  );
  port (
    clk          : in std_logic;
    reset_n      : in std_logic;
    enable       : in std_logic;
    m1           : in my_custom_matrix(0 to Np-1, 0 to Nm-1);
    m2           : in my_custom_matrix(0 to Nq-1, 0 to Nm-1);
    result_valid : out std_logic;
    result       : out dot_matrix_type(0 to Np-1, 0 to Nq-1)
  );
end entity;

architecture Behavioral of tile_array is

  component tile_unit
    generic (
      N  : integer;
      Nv : integer;
      Nh : integer;
      Nm : integer;
      Nd : integer
    );
    port (
      clk        : in std_logic;
      reset_n    : in std_logic;
      enable     : in std_logic;
      m1_chunk   : in my_custom_matrix(0 to (Np/Nt)-1, 0 to Nm-1);
      m2_chunk   : in my_custom_matrix(0 to Nh-1, 0 to Nm-1);
      out_valid  : out std_logic;
      result_out : out dot_matrix_type(0 to (Np/Nt)-1, 0 to Nh-1)
    );
  end component;

  type m1_tile_array is array(0 to Nt-1) of my_custom_matrix(0 to (Np/Nt)-1, 0 to Nm-1);
  type result_tile_array is array(0 to Nt-1) of dot_matrix_type(0 to (Np/Nt)-1, 0 to Nh-1);
  type result_global_array is array(0 to Nt-1, 0 to (Nq/Nh)-1) of dot_matrix_type(0 to (Np/Nt)-1, 0 to Nh-1);

  signal m1_tiles     : m1_tile_array;
  signal m2_slice     : my_custom_matrix(0 to Nh-1, 0 to Nm-1);
  signal tile_outputs : result_tile_array;
  signal tile_valids  : std_logic_vector(0 to Nt-1);
  signal global_result_chunks : result_global_array;

  signal current_m2_slice_idx : integer range 0 to (Nq/Nh);
  signal done_accumulating    : std_logic := '0';
  signal accumulating         : std_logic := '0';
  signal tile_enable          : std_logic := '0';

begin

  -- Split M1 into submatrices for each tile row
  m1_split_proc: process(m1)
  begin
    for t in 0 to Nt-1 loop
      for i in 0 to (Np/Nt)-1 loop
        for j in 0 to Nm-1 loop
          m1_tiles(t)(i,j) <= m1(t*(Np/Nt) + i, j);
        end loop;
      end loop;
    end loop;
  end process;

  -- Extract a slice from M2 to broadcast to all tiles
  m2_slice_proc : process(clk)
    variable all_tiles_done    : boolean;
    variable result_valid_var  : std_logic := '0'; 
  begin
    if rising_edge(clk) then
      if reset_n = '0' then
        current_m2_slice_idx <= 0;
        accumulating         <= '0';
        tile_enable          <= '0';
        done_accumulating    <= '0';
        result_valid_var     := '0';

      elsif enable = '1' and done_accumulating = '0' then
        if accumulating = '0' then
          for r in 0 to Nh-1 loop
            for k in 0 to Nm-1 loop
              m2_slice(r, k) <= m2(current_m2_slice_idx * Nh + r, k); 
            end loop;
          end loop;
          accumulating  <= '1';
          tile_enable   <= '1';

        else
          tile_enable <= '0'; 
          all_tiles_done := true;
          for i in 0 to Nt-1 loop
            if tile_valids(i) /= '1' then
              all_tiles_done := false;
            end if;
          end loop;

          if all_tiles_done then
            for t in 0 to Nt-1 loop
              global_result_chunks(t, current_m2_slice_idx) <= tile_outputs(t);
            end loop;

            if current_m2_slice_idx = (Nq / Nh) - 1 then
              done_accumulating <= '1';
              result_valid_var  := '1';  
            else
              current_m2_slice_idx <= current_m2_slice_idx + 1;
              accumulating         <= '0';
            end if;
          end if;
        end if;
      end if;

      result_valid <= result_valid_var;
      result_valid_var := '0';  
    end if;
  end process;

  -- Generate the tiles
  tile_gen: for t in 0 to Nt-1 generate
    tile_inst: tile_unit
      generic map (
        N  => N,
        Nv => (Np/Nt),
        Nh => Nh,
        Nm => Nm,
        Nd => Nd
      )
      port map (
        clk        => clk,
        reset_n    => reset_n,
        enable     => tile_enable,
        m1_chunk   => m1_tiles(t),
        m2_chunk   => m2_slice,
        out_valid  => tile_valids(t),
        result_out => tile_outputs(t)
      );
  end generate;

  -- Assemble final matrix from tile outputs
  result_assembly_proc: process(done_accumulating)
  begin
    if done_accumulating = '1' then
      for t in 0 to Nt-1 loop
        for chunk in 0 to (Nq/Nh)-1 loop
          for i in 0 to (Np/Nt)-1 loop
            for j in 0 to Nh-1 loop
              result(t*(Np/Nt) + i, chunk*Nh + j) <= global_result_chunks(t, chunk)(i,j);
            end loop;
          end loop;
        end loop;
      end loop;
    end if;
  end process;

end architecture;
