library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.generic_type_pkg.all;

entity tile_unit_tb is
end;

architecture sim of tile_unit_tb is
  constant Nv : integer := 2;
  constant Nh : integer := 2;
  constant Nm : integer := 4;
  constant Nd : integer := 2;
  constant N  : integer := 2;
  constant CUSTOM_WIDTH : integer := 16;

  constant clk_period : time := 10 ns;

  signal clk        : std_logic := '0';
  signal reset_n    : std_logic := '0';
  signal enable     : std_logic := '0';

  signal m1_chunk   : my_custom_matrix(0 to Nv-1, 0 to Nm-1);
  signal m2_chunk   : my_custom_matrix(0 to Nh-1, 0 to Nm-1);
  signal out_valid  : std_logic;
  signal result_out : dot_matrix_type(0 to Nv-1, 0 to Nh-1);

begin

  -------------------------------------------------------------------
  -- Clock Generation
  -------------------------------------------------------------------
  clk_process : process
  begin
    while now < 500 ns loop
      clk <= '0';
      wait for clk_period / 2;
      clk <= '1';
      wait for clk_period / 2;
    end loop;
    wait;
  end process;

  -------------------------------------------------------------------
  -- Unit Under Test
  -------------------------------------------------------------------
  uut: entity work.tile_unit
    generic map (
      N             => N,
      Nv            => Nv,
      Nh            => Nh,
      Nm            => Nm,
      Nd            => Nd,
      CUSTOM_WIDTH  => CUSTOM_WIDTH
    )
    port map (
      clk        => clk,
      reset_n    => reset_n,
      enable     => enable,
      m1_chunk   => m1_chunk,
      m2_chunk   => m2_chunk,
      out_valid  => out_valid,
      result_out => result_out
    );

  -------------------------------------------------------------------
  -- Stimulus Process
  -------------------------------------------------------------------
  stim_proc: process
  begin
    -- Reset and initial delay
    reset_n <= '0';
    enable  <= '0';
    wait for 2 * clk_period;
    reset_n <= '1';
    wait for clk_period;

    -- Initialize m1_chunk
    for i in 0 to Nv-1 loop
      for j in 0 to Nm-1 loop
        if i = j then
          m1_chunk(i, j) <= to_signed(2, CUSTOM_WIDTH);
        else
          m1_chunk(i, j) <= to_signed(1, CUSTOM_WIDTH);
        end if;
      end loop;
    end loop;

    -- Initialize m2_chunk with increasing numbers
    for i in 0 to Nh-1 loop
      for j in 0 to Nm-1 loop
        m2_chunk(i, j) <= to_signed(i * Nm + j + 1, CUSTOM_WIDTH);  -- e.g., 1,2,3,...
      end loop;
    end loop;

    -- Trigger computation
    enable <= '1';
    wait for 2 * clk_period;
    enable <= '0';

    -- Wait for valid output
    wait until out_valid = '1';

    report "Matrix multiplication result is valid.";

    -- Print result
    for i in 0 to Nv-1 loop
      for j in 0 to Nh-1 loop
        report "result_out(" & integer'image(i) & "," & integer'image(j) & ") = " &
               integer'image(to_integer(result_out(i,j)));
      end loop;
    end loop;

    wait;
  end process;

end architecture;
